import requests
import tarfile
import platform
import sys
import os


def download_tci(
    path=os.path.dirname(os.path.abspath(__file__)),
    version="8.4",
):
    os_platform = platform.system().lower()
    os_arch = "x64" if sys.maxsize > 2**32 else "x86_64"
    url = f"https://www.transaction.de/downloads/transbase/{version}/{os_platform}/{os_arch}/transbase_dev.tar.Z"
    print(f"xdownloading from {url} into {os.path.abspath(path)}")
    response = requests.get(url, stream=True)
    with tarfile.open(fileobj=response.raw, mode="r|gz") as tar:
        for member in tar:
            if member.name.find("tci") > -1:
                file = tar.extractfile(member)
                if file is None:
                    continue
                outfile = os.path.join(path, "lib", os.path.basename(member.name))
                os.makedirs(os.path.join(path, "lib"), exist_ok=True)
                with open(outfile, "wb") as output:
                    output.write(file.read())



import sys


class Error(Exception):
    """
    Exception that is the base class of all other error exceptions.
    You can use this to catch all errors with one single except statement.
    Warnings are not considered errors and thus should not use this class as base.
    It must be a subclass of the Python StandardError (defined in the module exceptions).
    """

    pass


class DatabaseError(Error):
    """
    Exception raised for errors that are related to the database.Subclass of Error.
    """

    def __init__(self, error_code, message="", sql_code=None):
        self.error_code = error_code
        self.traceback = sys.exc_info()
        sql = f"SQL-{sql_code}" if sql_code is not None else ""
        msg = f"TCI-{error_code}: {sql}: {message}"
        super().__init__(msg)


class InterfaceError(Error):
    """
    Exception raised for errors that are related to the database interface
    rather than the database itself. Subclass of Error.
    """

    pass



# https://www.python.org/dev/peps/pep-0249/
from collections import Mapping
from transbase import tci

# GLOBALS
apilevel = "2.0"
threadsafety = 0
paramstyle = "qmark,named"

State = tci.ResultState


class Cursor:
    """
    These objects represent a database cursor,
    which is used to manage the context of a fetch operation
    """

    """
    This read-only attribute is a sequence of 7-item sequences.
    Each of these sequences contains information describing one result column:
        - name
        - type_code
    (- display_size
        - internal_size
        - precision
        - scale
        - null_ok )
    """

    description = None
    """
    This read-only attribute specifies the number of rows that the last .execute*() produced
    """
    rowcount = 0
    """
    This read/write attribute specifies the number of rows to fetch at a time with .fetchmany().
    It defaults to 1 meaning to fetch a single row at a time
    """
    arraysize = 1

    """
    By default all column values are fetched as strings. Set type_cast to "True" to cast values
    to built in python data types where possible
    """
    type_cast = False

    __statement = tci.TCIEnvironment()
    __resultset = tci.TCIResultSet()

    def __init__(self, connection, error):
        self.__error = error
        self.__call(tci.allocateStatement(connection, error, self.__statement))
        self.__call(tci.allocateResultSet(self.__statement, error, self.__resultset))

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def execute(self, operation: str, params=None):
        """
        Prepare and execute a database operation (query or command).
        Parameters may be provided as sequence or mapping and will be bound to variables in the operation
        """
        # execute query direct if no parameters are given
        if params is None:
            self.__call(tci.executeDirect(self.__resultset, operation, 1, 0))

        # prepare statement and set positional or named parameters before execution
        else:
            self.__call(tci.prepare(self.__statement, operation))
            if isinstance(params, Mapping):
                for name, value in params.items():
                    self.__call(tci.set_param_by_name(self.__resultset, name, value))
            else:
                for idx, param in enumerate(params, start=1):
                    self.__call(tci.set_param(self.__resultset, idx, param))

            self.__call(tci.execute(self.__resultset, 1, 0))

        self.__setResultSetMeta()
        self.__setRowCount()

    def executemany(self, operation: str, seq_of_parameters=[]):
        """
        Prepare a database operation (query or command) and then execute it against all parameter sequences
        or mappings found in the sequence seq_of_parameters.
        """
        # for loop, optimize later
        # TODO
        pass

    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence, or None when no more data is available.
        """
        self.__state = State(tci.fetch(self.__resultset, 1, tci.TCI_FETCH_NEXT, 0))
        if self.__state == State.SUCCESS:
            return self.__getRow()
        elif self.__state != State.NO_DATA_FOUND:
            return self.__call(self.__state)
        else:
            return None

    def fetchmany(self, size=None):
        """
        Fetch the next set of rows of a query result, returning a sequence of sequences (e.g. a list of tuples).
        An empty sequence is returned when no more rows are available.
        Set size to -1 to fetch all remaining, see fetchall.
        """
        result = []
        count = size if size is not None else self.arraysize
        i = 0
        while count < 0 or i < count:
            i += 1
            # can be optimized later with tci fetch many...
            self.__state = State(tci.fetch(self.__resultset, 1, tci.TCI_FETCH_NEXT, 0))
            if self.__state == State.SUCCESS:
                result.append(self.__getRow())
            elif self.__state != State.NO_DATA_FOUND:
                self.__call(self.__state)
            else:
                break

        return result

    def fetchall(self):
        """
        Fetch all (remaining) rows of a query result, returning them as a sequence of sequences (e.g. a list of tuples).
        Note that the cursor's arraysize attribute can affect the performance of this operation.
        """
        return self.fetchmany(-1)

    def setinputsizes(self, sizes):
        """
        This can be used before a call to .execute*() to predefine memory areas for the operation's parameters.
        """
        # TODO (optional api)
        pass

    def setoutputsize(self, size, column=None):
        """
        Set a column buffer size for fetches of large columns (e.g. LONGs, BLOBs, etc.).
        The column is specified as an index into the result sequence. Not specifying the column will set the default size for all large columns in the cursor.
        """
        # TODO (optional api)
        pass

    def close(self):
        """
        Close the cursor now (rather than whenever __del__ is called)
        """
        if self.__resultset:
            tci.close(self.__resultset)
            tci.freeResultSet(self.__resultset)
            self.__resultset = None
        if self.__statement:
            tci.freeStatement(self.__statement)
            self.__statement = None
        self.__state = None
        self.description = None
        self.rowcount = 0
        self.arraysize = 1

    def state(self):
        return self.__state

    def __setResultSetMeta(self):
        no_cols = tci.resultset_attribute(self.__resultset, tci.TCI_ATTR_COLUMN_COUNT)
        meta = []
        for col in range(1, no_cols + 1):
            colName = tci.resultset_string_attribute(
                self.__resultset, tci.TCI_ATTR_COLUMN_NAME, col
            )
            colType = tci.resultset_attribute(
                self.__resultset, tci.TCI_ATTR_COLUMN_TYPE, col
            )
            meta.append([colName, colType])

        self.description = meta

    def __setRowCount(self):
        self.rowcount = tci.resultset_attribute(
            self.__resultset,
            tci.TCI_ATTR_ROWCOUNT,
        )

    def __getRow(self):
        row = []
        if self.description is None:
            return row
        for idx, info in enumerate(self.description, start=1):
            str_value = tci.get_data_as_string(self.__resultset, idx)
            row.append(self.__cast(info, str_value) if self.type_cast else str_value)
        return row

    def __call(self, tci_state):
        self.__state = State(tci_state)
        if self.__state != State.SUCCESS:
            tci.handle_error(self.__error)

    def __cast(self, info, value):
        sql_type = sql_type_code_to_name(info[1])
        if value is None:
            return None
        elif sql_type == "BOOL":
            return True if value.lower() == "true" else False
        elif sql_type in ["TINYINT", "SMALLINT", "INTEGER", "BIGINT"]:
            return int(value)
        elif sql_type in ["NUMERIC", "FLOAT", "DOUBLE"]:
            return float(value)
        # TODO fetch buffered and use
        # https://docs.python.org/3/c-api/buffer.html#bufferobjects
        elif sql_type in ["BLOB", "BINARY"]:
            return bytes.fromhex(value)
        else:
            return value

    # def callproc( procname [, parameters ] )


class Connection:
    __env = tci.TCIEnvironment()
    __con = tci.TCIConnection()
    __error = tci.TCIError()
    __tx = tci.TCITransaction()

    """
    Attribute to query and set the autocommit mode of the connection.
    Return True if the connection is operating in autocommit (non- transactional) mode.
    Return False if the connection is operating in manual commit (transactional) mode.
    """
    autocommit = True

    def setautocommit(self, value: bool):
        if self.autocommit and not value:
            self.__begin()
        self.autocommit = value

    def __init__(self, url, user, password, autocommit=True):
        self.autocommit = autocommit
        tci.allocateEnvironment(self.__env)
        tci.allocateError(self.__env, self.__error)
        self.__call(tci.allocateConnection(self.__env, self.__error, self.__con))
        self.__call(tci.allocateTransaction(self.__env, self.__error, self.__tx))
        self.__call(tci.connect(self.__con, url))
        self.__call(tci.login(self.__con, user, password))
        if not self.autocommit:
            self.__begin()

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def client_version(self):
        (state, version) = tci.client_version(self.__env)
        self.__call(state)
        return version

    def server_version(self):
        (state, version) = tci.server_version(self.__con)
        self.__call(state)
        return version

    def close(self):
        """
        Close the connection now (rather than whenever .__del__() is called).
        """
        if self.__tx:
            tci.freeTransaction(self.__tx)
            self.__tx = None
        if self.__con:
            tci.logout(self.__con)
            tci.disconnect(self.__con)
            tci.freeConnection(self.__con)
            self.__con = None
        if self.__error:
            tci.freeError(self.__error)
            self.__error = None
        if self.__env:
            tci.freeEnvironment(self.__env)
            self.__env = None
        self.__state = None

    def commit(self):
        """Commit any pending transaction to the database."""
        if self.__tx:
            self.__call(tci.commit(self.__tx))

    def rollback(self):
        """This method is optional since not all databases provide transaction support"""
        if self.__tx:
            self.__call(tci.rollback(self.__tx))

    def cursor(self) -> Cursor:
        """Return a new Cursor Object using the connection."""
        return Cursor(self.__con, self.__error)

    def state(self):
        return self.__state

    def __begin(self):
        """start a new transaction"""
        if self.__tx:
            self.__call(tci.beginTransaction(self.__tx, self.__con))

    def __call(self, tci_state):
        self.__state = State(tci_state)
        if self.__state != State.SUCCESS:
            tci.handle_error(self.__error)


# Constructor for creating a connection to the database
def connect(url: str, user: str, password: str) -> Connection:
    return Connection(url, user, password)


# get the sql type name for the given tci sql type code
def sql_type_code_to_name(code: int) -> str:
    return tci.TCI_SQL_TYPES_INVERTED.get(code) or "UNKNOWN"


def sql_type_name_to_code(name: str) -> int:
    return tci.TCI_SQL_TYPES.get(name) or 0



from setuptools import setup, Command
from transbase.install import download_tci
from pathlib import Path
import sys

TRANSBASE_VERSION = "8.4"


def get_lib_name():
    if sys.platform.startswith("win"):
        return "tci.dll"
    elif sys.platform.startswith("darwin"):
        return "libtci.dylib"
    else:
        return "libtci.so"


class DownloadSdkCommand(Command):
    """download tci sdk command."""

    description = "download tci sdk"
    user_options = []

    def initialize_options(self):
        # nothing todo
        pass

    def finalize_options(self):
        # nothing todo
        pass

    def run(self):
        download_tci(".", TRANSBASE_VERSION)


setup(
    name="transbase",
    packages=["transbase"],
    scripts=["scripts/install_tci"],
    has_ext_modules=lambda: True,
    package_data={
        "transbase": [f"../lib/{get_lib_name()}"],
    },
    version="0.0.6",
    description="Python Transbase Client",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://www.transaction.de",
    author="Daniel Loibl",
    author_email="loibl@transaction.de",
    license="MIT",
    test_suite="test",
    cmdclass={
        "get_lib": DownloadSdkCommand,
    },
    python_requires=">=3.9",
    install_requires=["requests"],
)



import ctypes as ct
from enum import Enum
import sys
from transbase.error import DatabaseError, Error
import os.path


def find_tci_lib():
    lib_name = None
    if sys.platform.startswith("win"):
        lib_name = "tci.dll"
    elif sys.platform.startswith("darwin"):
        lib_name = "libtci.dylib"
    else:
        lib_name = "libtci.so"

    packageDirectory = os.path.dirname(os.path.abspath(__file__)) + os.path.sep
    search_locations = [
        (
            packageDirectory + "lib" + os.path.sep + lib_name
        ),  # absolute site-packages/transbase/lib/ (e.g. when intalled with install_tci)
        (
            packageDirectory + ".." + os.path.sep + "lib" + os.path.sep + lib_name
        ),  # packaged whithin wheel ../lib
        (
            "." + os.path.sep + "lib" + os.path.sep + lib_name
        ),  # ./lib for local development, or if download manually there
    ]

    for path in search_locations:
        if os.path.isfile(path):
            return path
    raise Error(
        "\033[91m"
        + f"Unable to find tci sdk ({lib_name}) searched in {search_locations}.\n Please run 'install_tci' command."
        + "\033[0m"
    )


# load tci dll
if __name__ == "transbase.tci":
    tci = ct.CDLL(find_tci_lib())
else:
    raise Error("Exit")

sizeof = ct.sizeof

TCIEnvironment = ct.POINTER(ct.c_void_p)
TCIError = ct.POINTER(ct.c_void_p)
TCIConnection = ct.POINTER(ct.c_void_p)
TCIStatement = ct.POINTER(ct.c_void_p)
TCIResultSet = ct.POINTER(ct.c_void_p)
TCITransaction = ct.POINTER(ct.c_void_p)
TCIState = ct.c_int
TCIErrorCode = ct.c_int
attribute = ct.c_int


"""
ALLOCATIONS
"""
tci.TCIAllocEnvironmentW.argtypes = [ct.POINTER(TCIEnvironment)]
tci.TCIAllocEnvironmentW.restype = TCIState
allocateEnvironment = tci.TCIAllocEnvironmentW

tci.TCIAllocErrorW.argtypes = [TCIEnvironment, ct.POINTER(TCIError)]
tci.TCIAllocErrorW.restype = TCIState
allocateError = tci.TCIAllocErrorW

tci.TCIAllocConnectionW.argtypes = [TCIEnvironment, TCIError, ct.POINTER(TCIConnection)]
tci.TCIAllocConnectionW.restype = TCIState
allocateConnection = tci.TCIAllocConnectionW

tci.TCIAllocStatementW.argtypes = [TCIConnection, TCIError, ct.POINTER(TCIStatement)]
tci.TCIAllocStatementW.restype = TCIState
allocateStatement = tci.TCIAllocStatementW

tci.TCIAllocResultSetW.argtypes = [TCIStatement, TCIError, ct.POINTER(TCIResultSet)]
tci.TCIAllocResultSetW.restype = TCIState
allocateResultSet = tci.TCIAllocResultSetW

tci.TCIAllocTransactionW.argtypes = [
    TCIEnvironment,
    TCIError,
    ct.POINTER(TCITransaction),
]
tci.TCIAllocTransactionW.restype = TCIState
allocateTransaction = tci.TCIAllocTransactionW

"""
FUNCTIONS
"""
tci.TCIConnectW.argtypes = [TCIConnection, ct.c_wchar_p]
tci.TCIConnectW.restype = TCIState
connect = tci.TCIConnectW

tci.TCILoginW.argtypes = [TCIConnection, ct.c_wchar_p, ct.c_wchar_p]
tci.TCILoginW.restype = TCIState
login = tci.TCILoginW


tci.TCIGetErrorW.argtypes = [
    TCIError,
    ct.c_int,
    ct.c_int,
    ct.c_wchar_p,
    ct.c_int,
    ct.POINTER(TCIErrorCode),
    ct.c_wchar_p,
]
tci.TCIGetErrorW.restype = TCIState
getError = tci.TCIGetErrorW


def handle_error(error):
    message = ct.create_unicode_buffer("", 1024)
    code = TCIErrorCode()
    sql_code = ct.create_unicode_buffer("", 1024)
    getError(
        error,
        1,
        1,
        message,
        sizeof(message),
        code,
        sql_code,
    )
    raise DatabaseError(code.value, message.value, sql_code.value)


tci.TCIExecuteDirectW.argtypes = [TCIResultSet, ct.c_wchar_p, ct.c_int, ct.c_int]
tci.TCIExecuteDirectW.restype = TCIState
executeDirect = tci.TCIExecuteDirectW

tci.TCIExecuteW.argtypes = [TCIResultSet, ct.c_int, ct.c_int]
tci.TCIExecuteW.restype = TCIState
execute = tci.TCIExecuteW

tci.TCIPrepareW.argtypes = [TCIStatement, ct.c_wchar_p]
tci.TCIPrepareW.restype = TCIState
prepare = tci.TCIPrepareW

tci.TCIFetchW.argtypes = [TCIResultSet, ct.c_int, ct.c_int, ct.c_int]
tci.TCIFetchW.restype = TCIState
fetch = tci.TCIFetchW

tci.TCIBeginTransactionW.argtypes = [TCITransaction, TCIConnection]
tci.TCIBeginTransactionW.restype = TCIState
beginTransaction = tci.TCIBeginTransactionW

tci.TCICommitTransactionW.argtypes = [TCITransaction]
tci.TCICommitTransactionW.restype = TCIState
commit = tci.TCICommitTransactionW

tci.TCIRollbackTransactionW.argtypes = [TCITransaction]
tci.TCIRollbackTransactionW.restype = TCIState
rollback = tci.TCIRollbackTransactionW


def get_tci_type_and_c_value(value):
    if type(value) == int:
        (c_value, tci_type, by_ref) = (ct.c_long(value), TCI_C_INT8, True)
    elif type(value) == float:
        (c_value, tci_type, by_ref) = (ct.c_double(value), TCI_C_DOUBLE, True)
    elif type(value) == bool:
        (c_value, tci_type, by_ref) = (ct.c_bool(value), TCI_C_INT1, True)
    elif type(value) == bytearray or type(value) == bytes:
        (c_value, tci_type, by_ref) = (
            ct.create_string_buffer(value, len(value)),
            TCI_C_BYTE,
            False,
        )
    elif type(value) == str:
        (c_value, tci_type, by_ref) = (ct.c_wchar_p(value), TCI_C_WCHAR, False)
    else:
        (c_value, tci_type, by_ref) = (ct.c_wchar_p(str(value)), TCI_C_WCHAR, False)

    is_null = ct.c_int(-1) if value is None else None

    return (c_value, tci_type, by_ref, is_null)


def set_param(resultset, index, value):
    (c_value, tci_type, by_ref, is_null) = get_tci_type_and_c_value(value)
    return tci.TCISetDataW(
        resultset,
        index,
        c_value if not by_ref else ct.byref(c_value),
        sizeof(c_value),
        tci_type,
        is_null,
    )


def set_param_by_name(resultset, name, value):
    (c_value, tci_type, by_ref, is_null) = get_tci_type_and_c_value(value)
    return tci.TCISetDataByNameW(
        resultset,
        name,
        c_value if not by_ref else ct.byref(c_value),
        sizeof(c_value),
        tci_type,
        is_null,
    )


getData = tci.TCIGetDataW


def get_data_as_string(resultset, colNo=1):
    col = ct.c_int(colNo)
    byteSize = attribute(0)
    charLength = attribute(0)
    isNull = ct.c_bool()
    tci.TCIGetDataSizeW(resultset, col, TCI_C_WCHAR, ct.byref(byteSize), None)
    tci.TCIGetDataCharLengthW(resultset, col, ct.byref(charLength), None)
    buffer = ct.create_unicode_buffer(max(byteSize.value, charLength.value) + 1)
    getData(
        resultset,
        col,
        ct.byref(buffer),
        ct.sizeof(buffer),
        None,
        TCI_C_WCHAR,
        ct.byref(isNull),
    )
    return None if isNull else buffer.value


getResultSetAttribute = tci.TCIGetResultSetAttributeW


def resultset_attribute(resultset, attributekey, col=1):
    attr = attribute(0)
    tci.TCIGetResultSetAttributeW(
        resultset, attributekey, col, ct.byref(attr), sizeof(attr), None
    )
    return attr.value


def resultset_string_attribute(resultset, attributekey, col=1):
    attr = ct.create_unicode_buffer("", 127)
    tci.TCIGetResultSetAttributeW(
        resultset, attributekey, col, attr, sizeof(attr), None
    )
    return attr.value


class TCIVersion(ct.Structure):
    _fields_ = [
        ("major_version", ct.c_int),
        ("minor_version", ct.c_int),
        ("release", ct.c_int),
        ("patch", ct.c_int),
        ("build", ct.c_int),
    ]

    def __str__(self):
        return f"{self.major_version}.{self.minor_version}.{self.release}.{self.patch}.{self.build}"


def client_version(environment):
    version = TCIVersion()
    state = tci.TCIGetEnvironmentAttributeW(
        environment, TCI_ATTR_VERSION, 1, ct.byref(version), sizeof(version), None
    )
    return (state, version)


def server_version(connection):
    version = TCIVersion()
    state = tci.TCIGetConnectionAttributeW(
        connection, TCI_ATTR_VERSION, 1, ct.byref(version), sizeof(version), None
    )
    return (state, version)


"""
CLEANUP
"""
tci.TCIFreeResultSetW.argtypes = [TCIResultSet]
tci.TCIFreeResultSetW.restype = TCIState
freeResultSet = tci.TCIFreeResultSetW

tci.TCIFreeStatementW.argtypes = [TCIStatement]
tci.TCIFreeStatementW.restype = TCIState
freeStatement = tci.TCIFreeStatementW

tci.TCIFreeConnectionW.argtypes = [TCIConnection]
tci.TCIFreeConnectionW.restype = TCIState
freeConnection = tci.TCIFreeConnectionW

tci.TCIFreeErrorW.argtypes = [TCIError]
tci.TCIFreeErrorW.restype = TCIState
freeError = tci.TCIFreeErrorW

tci.TCIFreeEnvironmentW.argtypes = [TCIEnvironment]
tci.TCIFreeEnvironmentW.restype = TCIState
freeEnvironment = tci.TCIFreeEnvironmentW

tci.TCIFreeTransactionW.argtypes = [TCITransaction]
tci.TCIFreeTransactionW.restype = TCIState
freeTransaction = tci.TCIFreeTransactionW

tci.TCICloseW.argtypes = [TCIResultSet]
tci.TCICloseW.restype = TCIState
close = tci.TCICloseW

tci.TCILogoutW.argtypes = [TCIConnection]
tci.TCILogoutW.restype = TCIState
logout = tci.TCILogoutW

tci.TCIDisconnectW.argtypes = [TCIConnection]
tci.TCIDisconnectW.restype = TCIState
disconnect = tci.TCIDisconnectW
"""
CONSTANTS
"""


class ResultState(Enum):
    SUCCESS = 0
    ERROR = -1
    NO_DATA_FOUND = 100
    INVALID_HANDLE = 17001
    TRANSACTIONS_ROLLEDBACK = 17027
    DATA_TRUNCATION = 17031


TCI_FETCH_NEXT = 1

TCI_TYPE = 0x0100
TCI_C_CHAR = TCI_TYPE | 0x1000 | 0x0A
TCI_C_WCHAR = TCI_C_CHAR | (0x2000)
TCI_C_INT1 = TCI_TYPE | 0x01
TCI_C_INT_4 = TCI_TYPE | 0x03
TCI_C_INT8 = TCI_TYPE | 0x0B
TCI_C_DOUBLE = TCI_TYPE | 0x06
TCI_C_BYTE = TCI_TYPE | 0x1000 | 0x0C

TCI_ATTR_COLUMN_COUNT = 7
TCI_ATTR_COLUMN_NAME = 10
TCI_ATTR_COLUMN_TYPE = 14
TCI_ATTR_ROWCOUNT = 40
TCI_ATTR_RECORDS_TOUCHED = 55
TCI_ATTR_VERSION = 68

TCI_SQL_TYPES = {
    "BOOL": 16641,
    "TINYINT": 16642,
    "SMALLINT": 16643,
    "INTEGER": 16644,
    "NUMERIC": 16645,
    "FLOAT": 16646,
    "DOUBLE": 16647,
    "CHAR": 16648,
    "VARCHAR": 16649,
    "BINARY": 16650,
    "BIT": 16651,
    "BLOB": 16652,
    "BITSHORT": 16654,
    "BIGINT": 16655,
    "CLOB": 16656,
    "DATE": 17781,
    "DATE_YEAR": 17783,
    "DATE_YEAR_TO_MONTH": 17782,
    "DATE_YEAR_TO_DAY": 17781,
    "DATE_YEAR_TO_HOUR": 17780,
    "DATE_YEAR_TO_MINUTE": 17779,
    "DATE_YEAR_TO_SECOND": 17778,
    "DATE_YEAR_TO_MILLISECOND": 17777,
    "DATE_MONTH": 17766,
    "DATE_MONTH_TO_DAY": 17765,
    "DATE_MONTH_TO_HOUR": 17764,
    "DATE_MONTH_TO_MINUTE": 17763,
    "DATE_MONTH_TO_SECOND": 17762,
    "DATE_MONTH_TO_MILLISECOND": 17761,
    "DATE_DAY": 17749,
    "DATE_DAY_TO_HOUR": 17748,
    "DATE_DAY_TO_MINUTE": 17747,
    "DATE_DAY_TO_SECOND": 17746,
    "DATE_DAY_TO_MILLISECOND": 17745,
    "DATE_HOUR": 17732,
    "DATE_HOUR_TO_MINUTE": 17731,
    "DATE_HOUR_TO_SECOND": 17730,
    "DATE_HOUR_TO_MILLISECOND": 17729,
    "DATE_MINUTE": 17715,
    "DATE_MINUTE_TO_SECOND": 17714,
    "DATE_MINUTE_TO_MILLISECOND": 17713,
    "DATE_SECOND": 17698,
    "DATE_SECOND_TO_MILLISECOND": 17697,
    "DATE_MILLISECOND": 17681,
    "TIME": 17730,
    "TIMESTAMP": 17777,
    "INTERVAL": 18688,
    "INTERVAL_YEAR": 18807,
    "INTERVAL_YEAR_TO_MONTH": 18806,
    "INTERVAL_MONTH": 18790,
    "INTERVAL_DAY": 18773,
    "INTERVAL_DAY_TO_HOUR": 18772,
    "INTERVAL_DAY_TO_MINUTE": 18771,
    "INTERVAL_DAY_TO_SECOND": 18770,
    "INTERVAL_DAY_TO_MILLISECOND": 18769,
    "INTERVAL_HOUR": 18756,
    "INTERVAL_HOUR_TO_MINUTE": 18755,
    "INTERVAL_HOUR_TO_SECOND": 18754,
    "INTERVAL_HOUR_TO_MILLISECOND": 18753,
    "INTERVAL_MINUTE": 18739,
    "INTERVAL_MINUTE_TO_SECOND": 18738,
    "INTERVAL_MINUTE_TO_MILLISECOND": 18737,
    "INTERVAL_SECOND": 18722,
    "INTERVAL_SECOND_TO_MILLISECOND": 18721,
    "INTERVAL_MILLISECOND": 18705,
}
TCI_SQL_TYPES_INVERTED = {v: k for k, v in TCI_SQL_TYPES.items()}
